
#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import ssl
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.tools.tools import Tools
from holado_json.ipc.json import set_object_attributes_with_json_dict
from holado_core.common.exceptions.technical_exception import TechnicalException
import os
import copy
from ssl import Purpose

logger = logging.getLogger(__name__)


class SslManager(object):
    """
    Helper for ssl module.
    
    It is implemented internally with standard library ssl.
    """
    
    @classmethod
    def new_ssl_context(cls, server_side=False, flat_kwargs=None):
        res = None
        
        if flat_kwargs is None:
            flat_kwargs = {}
        kwargs = copy.copy(flat_kwargs)
        
        try:
            activate_ssl = kwargs.pop('activate', True)
            if activate_ssl:
                purpose = Purpose.CLIENT_AUTH if server_side else Purpose.SERVER_AUTH
                res = ssl.create_default_context(purpose=purpose)
                
                if Tools.has_sub_kwargs(kwargs, 'context.'):
                    context_kwargs = Tools.pop_sub_kwargs(kwargs, 'context.')
                    
                    ciphers = context_kwargs.pop('ciphers', None)
                    if ciphers is not None:
                        res.set_ciphers(ciphers)
                    if Tools.has_sub_kwargs(context_kwargs, 'load_cert_chain.'):
                        load_cert_chain_kwargs = Tools.pop_sub_kwargs(context_kwargs, 'load_cert_chain.')
                        res.load_cert_chain(**load_cert_chain_kwargs)
                    if Tools.has_sub_kwargs(context_kwargs, 'load_verify_locations.'):
                        load_verify_locations_kwargs = Tools.pop_sub_kwargs(context_kwargs, 'load_verify_locations.')
                        res.load_verify_locations(**load_verify_locations_kwargs)
                    
                    # Set context attributes with remaining kwargs
                    if len(context_kwargs) > 0:
                        set_object_attributes_with_json_dict(res, context_kwargs)
        except Exception as exc:
            msg = f"Failed to create SSL context with parameters {flat_kwargs}: {exc}"
            logger.error(msg)
            raise TechnicalException(msg) from exc
        
        # Verify all kwargs were applied
        if len(kwargs) > 0:
            raise TechnicalException(f"Unmanaged ssl parameters: {kwargs}")
        
        return res
    
    @classmethod
    def get_localhost_certificate(cls):
        here = os.path.abspath(os.path.dirname(__file__))
        certfile_path = os.path.join(here, 'resources', 'certificates', 'localhost.crt')
        keyfile_path = os.path.join(here, 'resources', 'certificates', 'localhost.key')
        return (certfile_path, keyfile_path)





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado.common.handlers.object import DeleteableObject
import os
from typing import AnyStr, List


class File(DeleteableObject):
    """
    Manage a file
    """

    def __init__(self, path, **open_kwargs):
        super().__init__(f"file '{path}'")

        self.__path = path
        self.__auto_flush = True
        self.__file = None
        
        if open_kwargs:
            self.open(**open_kwargs)

    def _delete_object(self):
        self.close()
        
    @property
    def path(self):
        return self.__path

    @property
    def internal_file(self):
        return self.__file

    def open(self, auto_flush=True, **kwargs):
        self.__auto_flush = auto_flush
        self.__file = open(self.__path, **kwargs)

    def close(self):
        if self.__file:
            self.__file.close()
            self.__file = None

    def write(self, data):
        res = self.internal_file.write(data)
        if self.__auto_flush:
            self.internal_file.flush()
        return res

    def writelines(self, lines, add_line_sep=True):
        for line in lines:
            self.internal_file.write(line)
            if add_line_sep and not line.endswith(os.linesep):
                self.internal_file.write(os.linesep)
        if self.__auto_flush:
            self.internal_file.flush()

    def read(self, n: int = -1) -> AnyStr:
        return self.internal_file.read(n)

    def readline(self, limit: int = -1) -> AnyStr:
        return self.internal_file.readline(limit)

    def readlines(self, hint: int = -1) -> List[AnyStr]:
        return self.internal_file.readlines(hint)





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_value.common.tables.value_table_with_header import ValueTableWithHeader
from holado_core.common.tables.table_manager import TableManager
from holado_core.common.tools.tools import Tools
from holado_value.common.tables.value_table import ValueTable
from holado_core.common.exceptions.verify_exception import VerifyException

logger = logging.getLogger(__name__)


class ValueTableManager(TableManager):
    
    @classmethod
    def is_value_table(cls, table):
        return isinstance(table, ValueTable) or isinstance(table, ValueTableWithHeader)
    
    @classmethod
    def verify_is_value_table(cls, table, raise_exception=True):
        res = cls.is_value_table(table)
        
        if not res and raise_exception:
            raise VerifyException(f"Table is not a ValueTable of ValueTableWithHeader (table type: {type(table)})")
        return res
    
    @classmethod
    def convert_table_2_list_of_tuples(cls, table):
        if isinstance(table, ValueTable) or isinstance(table, ValueTableWithHeader):
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug("Converting value table to a list of tuples...")
            return [tuple((cell.value for cell in row.cells)) for row in table.rows]
        else:
            return TableManager.convert_table_2_list_of_tuples(table)
    
    
    @classmethod
    def set_object_attributes_according_name_value_table(cls, obj, table):
        if not cls.is_value_table(table):
            super().set_object_attributes_according_name_value_table(obj, table)
        
        # Verify table structure
        cls.verify_table_is_name_value_table(table)
        
        for row in table.rows:
            setattr(obj, row.get_cell(0).value, row.get_cell(1).value)
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import abc


logger = logging.getLogger(__name__)



class QueryBuilder():
    """
    Generic Query builder.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, name, db_client=None):
        self.__name = name
        self.__db_client = db_client
    
    @property
    def name(self):
        return self.__name
    
    @property
    def db_client(self):
        return self.__db_client
    
    @db_client.setter
    def db_client(self, client):
        self.__db_client = client
    
    def select(self, table_name, where_data: dict=None, sql_return="*"):
        """
        Simple query & values builder of a select by filtering on given where data.
        Parameter 'where_data' has to be a dictionary with keys equal to table column names.
        """
        raise NotImplementedError

    def insert(self, table_name, data: dict):
        """
        Simple query & values builder of an insert of given data.
        Parameter 'data' has to be a dictionary with keys equal to table column names.
        """
        raise NotImplementedError
    
    def update(self, table_name, data: dict, where_data: dict):
        """
        Simple query & values builder of an update of given data.
        Parameters 'data' and 'where_data' have to be dictionaries with keys equal to table column names.
        """
        raise NotImplementedError
    
    def delete(self, table_name, where_data: dict=None):
        """
        Simple query & values builder of a delete by filtering on given where data.
        Parameter 'where_data' has to be a dictionary with keys equal to table column names.
        """
        raise NotImplementedError
    
    def where(self, query, values, where_data: dict):
        """
        Add where clauses to current couple (query, values), and return a new couple (query, values).
        """
        raise NotImplementedError
    
    def where_in(self, query, values, field_name, field_values, not_in=False):
        """
        Add where in clause to current couple (query, values), and return a new couple (query, values).
        """
        raise NotImplementedError
        
    
    def to_sql(self, query):
        raise NotImplementedError
        



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2023 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from builtins import NotImplementedError
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.tables.table_with_header import TableWithHeader
from holado_core.common.tables.table_row import TableRow
import abc
from holado_core.common.tools.tools import Tools
from holado_core.common.tables.table import Table
from holado_db.tools.db.query.base.query_builder import QueryBuilder

logger = logging.getLogger(__name__)


class DBClient(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self, name, connect_kwargs, with_cursor=True, auto_commit=True):
        self.__name = name
        self.__connect_kwargs = connect_kwargs
        self.__with_cursor = with_cursor
        self.__auto_commit = auto_commit
        
        self.__connection = None
        self.__query_builder = None
    
    @property
    def name(self):
        return self.__name
    
    @property
    def query_builder(self) -> QueryBuilder:
        return self.__query_builder
    
    @query_builder.setter
    def query_builder(self, builder):
        self.__query_builder = builder
    
    @property
    def is_connected(self):
        return self.__connection is not None
    
    @property
    def connection(self):
        return self.__connection
    
    @property
    def cursor(self):
        if not self.__with_cursor:
            raise TechnicalException(f"DB client '{self.__name}' doesn't manage cursor")
        return self.__cursor
    
    def connect(self):
        try:
            self.__connection = self._connect(**self.__connect_kwargs)
        except Exception as exc:
            Tools.raise_same_exception_type(exc, f"[{self.name}] Failed to connect with parameters {self.__connect_kwargs}")
        if self.__with_cursor:
            self.__cursor = self.__connection.cursor()
    
    def _connect(self, **kwargs):
        raise NotImplementedError()
    
    def _verify_is_connected(self):
        if not self.is_connected:
            raise FunctionalException(f"DB Client '{self.name}' is not connected")
        
    def execute_query(self, query, *args, **kwargs):
        sql = self.query_builder.to_sql(query)
        return self.execute(sql, *args, **kwargs)
        
    def execute(self, sql, *args, **kwargs):
        # Manage commit & auto commit
        do_commit = None
        if 'do_commit' in kwargs:
            do_commit = kwargs.pop('do_commit')
        if do_commit is None:
            do_commit = self.__auto_commit
            
        self._verify_is_connected()
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] Executing SQL [{sql}] with parameters [{args if args else ''}{kwargs if kwargs else ''}]...")
        try:
            if args:
                self.cursor.execute(sql, args)
            elif kwargs:
                self.cursor.execute(sql, kwargs)
            else:
                self.cursor.execute(sql)
        except self._get_base_exception_type() as exc:
            self.rollback()
            raise TechnicalException(f"[{self.name}] Error while executing SQL [{sql}] (args: {args} ; kwargs: {kwargs})") from exc
        
        if self.cursor.description:
            field_names = [field[0] for field in self.cursor.description]
            
            res = TableWithHeader()
            res.header = TableRow(cells_content=field_names)
            
            row_values = self.cursor.fetchone()
            while row_values:
                res.add_row(cells_content=row_values)
                row_values = self.cursor.fetchone()
        elif self.cursor.rowcount > 0:
            res = self.cursor.rowcount
        else:
            res = None
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] Executed SQL [{sql}] with parameters [{args if args else ''}{kwargs if kwargs else ''}] => [{res}]")
        self.__log_sql_result(res)
        
        # Manage commit
        if do_commit:
            try:
                self.commit()
            except self._get_base_exception_type() as exc:
                self.rollback()
                raise TechnicalException(f"[{self.name}] Error while commit after SQL [{sql}] (args: {args} ; kwargs: {kwargs})") from exc
            
        return res
    
    def _get_base_exception_type(self):
        raise NotImplementedError()
    
    def __log_sql_result(self, sql_result, limit_rows=10):
        if Tools.do_log(logger, logging.DEBUG):
            res_str = self.__represent_sql_result(sql_result, limit_rows=limit_rows)
            if '\n' in res_str:
                logger.debug(f"[{self.name}] SQL result:\n{Tools.indent_string(4, res_str)}")
            else:
                logger.debug(f"[{self.name}] SQL result: {res_str}")
    
    def __represent_sql_result(self, sql_result, limit_rows = 10):
        if isinstance(sql_result, Table):
            return sql_result.represent(limit_rows=limit_rows)
        else:
            return str(sql_result)
    
    def commit(self):
        self.connection.commit()
        
    def rollback(self):
        self.connection.rollback()
        
    def exist_table(self, table_name):
        raise NotImplementedError()
    
    def insert(self, table_name, data: dict, do_commit=None):
        """
        Insert given data.
        Parameter 'data' has to be a dictionary with keys equal to table column names.
        """
        query, values = self.query_builder.insert(table_name, data)
        return self.execute_query(query, *values, do_commit=do_commit)
    
    def update(self, table_name, data: dict, where_data: dict, do_commit=None):
        """
        Update given data.
        Parameters 'data' and 'where_data' have to be dictionaries with keys equal to table column names.
        """
        query, values = self.query_builder.update(table_name, data, where_data)
        return self.execute_query(query, *values, do_commit=do_commit)
    
    def select(self, table_name, where_data: dict=None, sql_return="*"):
        """
        Select by filtering on given where data.
        Parameter 'where_data' has to be a dictionary with keys equal to table column names.
        """
        query, values = self.query_builder.select(table_name, where_data, sql_return)
        return self.execute_query(query, *values, do_commit=False)
    
    def delete(self, table_name, where_data: dict=None, do_commit=None):
        """
        Delete by filtering on given where data.
        Parameter 'where_data' has to be a dictionary with keys equal to table column names.
        """
        query, values = self.query_builder.delete(table_name, where_data)
        return self.execute_query(query, *values, do_commit=do_commit)
    
    def set_or_update_json_key_value(self, table_name, field_name, json_key, json_value, where_data):
        """
        Set or update a json field with key=value.
        """
        raise NotImplementedError()
        
    def _get_sql_placeholder(self):
        """
        Return the character/string to use as placeholder in SQL requests.
        """
        raise NotImplementedError()




# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


from holado_test.scenario.step_tools import StepTools
from holado.common.context.session_context import SessionContext
from holado_test.behave.behave import *
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_rest.api.rest.rest_client import RestClient
import json
import logging
from holado_rest.api.rest.rest_manager import RestManager
from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
from holado_core.common.tables.table_manager import TableManager
from holado_value.common.tables.converters.value_table_converter import ValueTableConverter
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)


if RestClient.is_available():

    def __get_session_context():
        return SessionContext.instance()
    
    def __get_scenario_context():
        return __get_session_context().get_scenario_context()
    
    def __get_text_interpreter():
        return __get_scenario_context().get_text_interpreter()
    
    def __get_variable_manager():
        return __get_scenario_context().get_variable_manager()
    
        
    @Given(r"(?P<var_name>{Variable}) = new REST client")
    def step_impl(context, var_name):
        var_name = StepTools.evaluate_variable_name(var_name)
        table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
        kwargs = ValueTableConverter.convert_name_value_table_2_dict(table)
        
        if 'name' not in kwargs:
            kwargs['name'] = None
        
        manager = RestManager()
        client = manager.new_client(**kwargs)
    
        __get_variable_manager().register_variable(var_name, client)
    
    @Step(r"(?P<var_name>{Variable}) = response of (?P<request_method>delete|get|patch|post|put) (?P<path>{Str}) \(REST client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_name, request_method, path, var_client):
        var_name = StepTools.evaluate_variable_name(var_name)
        path = StepTools.evaluate_scenario_parameter(path)
        client = StepTools.evaluate_variable_value(var_client)

        # Manage data as table
        data = None
        if hasattr(context, "table") and context.table is not None:
            table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
            data_obj = ValueTableConverter.convert_name_value_table_2_json_object(table)
            data = json.dumps(data_obj)
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"Transformed parameters to JSON is: {data}")
            
        # Manage data as json text
        text = BehaveStepTools.get_step_multiline_text(context, raise_exception_if_none=False)
        if text is not None:
            if request_method in ['delete', 'get']:
                raise TechnicalException(f"Unexpected json text for request method '{request_method}'")
            if data is not None:
                raise TechnicalException(f"Data is already set to [{data}]")
            data = text
        
        # Process request
        res = client.request(request_method, path, data=data)
        
        # Store result
        if Tools.do_log(logger, logging.DEBUG):
            res_table = TableManager.convert_object_attributes_2_name_value_table(res)
            logger.debug(f"Response of {request_method} {path}:\n{res_table.represent(4)}")
        __get_variable_manager().register_variable(var_name, res)
    
    @When(r"(?P<var_name>{Variable}) = (?P<request_method>delete|get|patch|post|put) (?P<path>{Str}) \(REST client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_name, request_method, path, var_client):
        var_name = StepTools.evaluate_variable_name(var_name)
        client = StepTools.evaluate_variable_value(var_client)

        execute_steps(format_step_with_context(context, 
                u"__RESPONSE__@ = response of {request_method} {path} (REST client: {var_client})"
                    .format(request_method=request_method, path=path, var_client=var_client) ))
        
        response = __get_variable_manager().get_variable_value("__RESPONSE__@")
        res = client.response_result(response)
        
        __get_variable_manager().register_variable(var_name, res)
    
    
    @Step(r"(?P<var_name>{Variable}) = response of (?P<request_method>delete|get|patch|post|put) (?P<path>{Str}) with content of CSV file (?P<file_path>{Str}) \(REST client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_name, request_method, path, file_path, var_client):
        var_name = StepTools.evaluate_variable_name(var_name)
        path = StepTools.evaluate_scenario_parameter(path)
        file_path = StepTools.evaluate_scenario_parameter(file_path)
        client = StepTools.evaluate_variable_value(var_client)
        
        # Request with file object as data
        with open(file_path, "rb") as data:
            res = client.request(request_method, path, data=data, headers={'Content-Type':'text/csv'})
        
        # Store result
        if Tools.do_log(logger, logging.DEBUG):
            res_table = TableManager.convert_object_attributes_2_name_value_table(res)
            logger.debug(f"Response of {request_method} {path}:\n{res_table.represent(4)}")
        __get_variable_manager().register_variable(var_name, res)
    
    @When(r"(?P<var_name>{Variable}) = (?P<request_method>delete|get|patch|post|put) (?P<path>{Str}) with content of CSV file (?P<file_path>{Str}) \(REST client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_name, request_method, path, file_path, var_client):
        var_name = StepTools.evaluate_variable_name(var_name)
        client = StepTools.evaluate_variable_value(var_client)

        execute_steps(format_step_with_context(context, 
                u"__RESPONSE__@ = response of {request_method} {path} with content of CSV file {file_path} (REST client: {var_client})"
                    .format(request_method=request_method, path=path, file_path=file_path, var_client=var_client) ))
        
        response = __get_variable_manager().get_variable_value("__RESPONSE__@")
        res = client.response_result(response)
        
        __get_variable_manager().register_variable(var_name, res)
    
    
    
